(***********************************************************************)
(*                  absyn.n                                            *)	
(*---------------------------------------------------------------------*)
(*  Initiatial development by Eun-sun Cho                              *)
(*  Finalized by Hyun-Goo Kang, ROPAS                                  *)
(*---------------------------------------------------------------------*)
(* 1. define a abstract syntax tree of XML document                    *)
(* 2. pretty printing                                                  *)
(*---------------------------------------------------------------------*)
(*    -- todo  --                                                      *)
(*---------------------------------------------------------------------*)
(*  should implement generation of DOM like structure using this absyn *)
(***********************************************************************)

structure Nxml_ast  = struct 

type t_pi = PI_WITH_TARGET of string * string
          | PI_WO_TARGET of string
type t_marks = STAR | PLUS | QMARK | NO_MARK 
type t_cp = t_pre_cp * t_marks
and t_pre_cp = NAME_CP of string 
             | CHOICE_CP of t_cp list
             | SEQ_CP of t_cp list
             | PEREF_CP of string
type t_pre_children = SEQ of t_cp list | CHOICE of t_cp list
type t_mixed =  MIXED_ONLY | MIXED_STAR of string list 
type t_content_spec = EMPTY_CS 
                | ANY_CS 
                | CHILDREN of t_pre_children * t_marks 
                | MIXED of t_mixed
type t_external_id = SYSTEM of string 
                | PUBLIC of string * string
type t_opt_ndata = EMPTY_NDATA | NDATA of string
type t_entity_def = GSTR_VAL of string 
                | EXTERNAL_GENT of (t_external_id * t_opt_ndata)
type t_pe_def =  PSTR_VAL of string 
                | EXTERNAL_PENT of t_external_id 
type t_entity_decl = GEDECL of  (string * t_entity_def)
                | PEDECL of (string  * t_pe_def)
type t_default_decl = REQUIRED | IMPLIED 
	        | FIXED of string | ATTVALUE of string 
type t_att_type = CDATA |  ID | IDREF | IDREFS | ENTITY
		| ENTITIES | NMTOKEN | NMTOKENS 		
		| NOTATION of string list
		| ENUMERATION of string list
type t_post_notation_decl =  EXTERNAL_ID_NOTATION of t_external_id
                | PUBID of string
type t_markup_decl = ELEMENT_DECL  of string * t_content_spec 
        	| ATTLIST_DECL of string * ( (string * t_att_type * t_default_decl)  list)
        	| ENTITY_DECL of t_entity_decl
        	| NOTATION_DECL of string * t_post_notation_decl
        	| PI_MARKUP of t_pi
        	| COMMENT_MARKUP of string
type t_internal_decl_body =  MARKUP of t_markup_decl 
type t_opt_internal_decl_list = EMPTY_INTERNAL_DECL | INTERNAL_DECL_LIST of (t_internal_decl_body list)

type t_include = INCLUDE | IGNORE
type t_ext_subset_decl_elt  =  MARKUP_EXT of t_markup_decl
	        | CONDITIONALSECT of t_include * t_ext_subset_decl
	        | PEREF_EXT of string
 and t_ext_subset_decl  =  t_ext_subset_decl_elt list

type t_misc =  PI of t_pi | COMMENT of string
type t_opt_standalone = EMPTY_STANDALONE | NO_STANDALONE | YES_STANDALONE
type t_opt_encoding = EMPTY_ENCODING | ENCODING of string
type t_opt_version = EMPTY_VERSION | VERSION of string
type t_opt_xml_decl = EMPTY_XML_DECL 
	        | XML_DECL of t_opt_version * t_opt_encoding * t_opt_standalone
type t_opt_text_decl = EMPTY_TEXTDECL 
		| TEXTDECL of string
		| TEXTDECL_WITHVERSION of string * string
type t_opt_external_id = EMPTY_EXTERNALID 
                | EXTERNALID of t_external_id * t_entry_point
 and t_doctype_decl = string *  		(* Doctype name *)
		t_opt_external_id * t_opt_internal_decl_list
 and t_opt_doctype_decl_miscs = EMPTY_ODTDM 
        	| ODTDM of  t_doctype_decl * (t_misc list)
 and t_prolog = t_opt_xml_decl  * (t_misc list) * t_opt_doctype_decl_miscs
 and t_content =  ELEMENT_CNT of t_element 
                | CHARDATA_CNT of string 
        	//| EREF_CNT of string 
        	| CDATA_CNT of string 
        	| PI_CNT of t_pi 
        	| COMMENT_CNT of string
 and t_element =  EMPTY_ELEMENT_TAG of string * ((string * string) list)
		| FULL_ELEMENT of 
	        	string * ((string * string) list) * t_content list * string 
 and t_entry_point = DOCUMENT of t_prolog * t_element * (t_misc list)
                | EXT_PARSEDENT of t_opt_text_decl * (t_content list)
                | EXT_PE of t_opt_text_decl * t_ext_subset_decl


(*************************)
(*  utilities for print  *)
(*************************)
(* list *)
fun print_list fnn nil p = ()
  | print_list fnn (s::nil) p = (fnn s) 
  | print_list fnn (s::alist) p = (fnn s; print_string p; print_list fnn alist p) 	
(* string surrounded by quotation marks *)
fun print_quotes s =  print_string ("\""^s^"\"")
(* pair *)
fun printpair pr1 pr2 ch (s1, s2) = (pr1 s1; print_string ch; pr2 s2)
(* quote *)
fun quote() = print_string "\""


(******************************)
(* print each node of the ast *)
(******************************)

(* external_id *)
fun print_external_id (SYSTEM(s1)) = 
	(print_string "SYSTEM "; print_quotes s1)
  | print_external_id (PUBLIC (s2, s3)) = 
	(print_string "PUBLIC "; 
	print_quotes s2; print_string " "; print_quotes s3)

(* PI *)
fun print_pi (PI_WO_TARGET t) = 
	print_string ("<?" ^ t ^ "?>\n")
  | print_pi (PI_WITH_TARGET (t1, t2)) = 
	print_string ("<?" ^ t1 ^" "^ t2 ^ "?>\n")

(* misc *)
fun print_misc (PI t) = print_pi t
  | print_misc (COMMENT  s) = (print_string (s^"\n"))

(* textdecl *)
fun print_textdecl EMPTY_TEXTDECL =  ()
  | print_textdecl (TEXTDECL s1 )  = 
	(print_string "<?xml encoding = ";
	print_quotes s1;
	print_string "?>\n")
  | print_textdecl (TEXTDECL_WITHVERSION (s2, s3) )  = 
	(print_string "<?xml version = ";
	print_quotes s2;
	print_string " encoding = ";
	print_quotes s3;
	print_string "?>\n")

(* XML decleration *)
fun print_xml_decl EMPTY_XML_DECL = ()
  | print_xml_decl (XML_DECL (v, e, s)) =
	(
	print_string "<?xml ";
	(case v of
		EMPTY_VERSION => ()
		| VERSION v =>  (print_string "version  ="; print_quotes(v))
	);
	print_string " ";
	(case e of
		EMPTY_ENCODING => ()
		| ENCODING e => (print_string "encoding ="; print_quotes(e))
	);
	print_string " ";
	(case s of
		EMPTY_STANDALONE => ()
		| NO_STANDALONE  =>  print_string "standalone = \"yes\""
		| YES_STANDALONE  =>  print_string "standalone = \"no\"");
	print_string "?>\n"
	)

(* attribute type *)
fun print_atttype CDATA  =  print_string "CDATA"
  | print_atttype ID  = print_string "ID"
  | print_atttype IDREF  =  print_string "IDREF"
  | print_atttype IDREFS =  print_string "IDREFS"
  | print_atttype ENTITY = print_string "ENTITY"
  | print_atttype ENTITIES  = print_string "ENTITIES"
  | print_atttype NMTOKEN  = print_string "NMTOKEN"
  | print_atttype NMTOKENS =print_string "NMTOKENS"
  | print_atttype (NOTATION sl1) = 
	(print_string "NOTATION (";
	print_list print_string sl1 "|"; 
	print_string ")")
  | print_atttype (ENUMERATION sl2) = 
	(print_string "(";
	print_list print_string sl2 "|"; 
	print_string ")")

(* *, +, ? *)
fun print_marks STAR = print_string "*" 
  | print_marks PLUS = print_string "+" 
  | print_marks QMARK = print_string "?" 
  | print_marks NO_MARK = ()

(* cp *)
fun print_cp (pcp, mark) = (print_pre_cp pcp ; print_marks mark)
and print_pre_cp (NAME_CP s) = print_string s
  | print_pre_cp (CHOICE_CP t) = 
	(print_string "(";	
	 print_list print_cp t "|";
	 print_string ")" )
  | print_pre_cp (SEQ_CP t) = 
	(print_string "("; 
	print_list print_cp t ",";
	print_string ")")
  | print_pre_cp (PEREF_CP t) = print_string ("%"^t^";")

(* content spec *)
fun print_content_spec EMPTY_CS  = print_string "EMPTY"
  | print_content_spec ANY_CS = print_string "ANY"
  | print_content_spec (MIXED MIXED_ONLY) = 
	(print_string "(#PCDATA)")
  | print_content_spec (MIXED (MIXED_STAR slist)) = 
	(print_string "(#PCDATA "; 
	 (case slist of
		nil => print_string ")"
		| t => ( print_string "|"; print_list print_string t "|" ; print_string ")*")
	 )
	)
  | print_content_spec (CHILDREN (x, opt)) = 
	( (case x of
		SEQ t => (print_string "("; 
			print_list print_cp t ",";
			print_string ")")
		| CHOICE t => (print_string "(";
			print_list print_cp t "|";
			print_string ")" )
	  );
	 print_marks opt
	)

(* default declaration *)
fun print_default_decl REQUIRED = print_string "#REQUIRED"
  | print_default_decl IMPLIED = print_string "#IMPLIED"
  | print_default_decl (FIXED s) = (print_string "#FIXED ";
				  print_quotes s)	
  | print_default_decl (ATTVALUE ss) = print_quotes ss

(* one attribute declaration *)
fun print_att (s, ty, dd) = 
	(print_string (s^" ");
	print_atttype ty; print_string " ";
	print_default_decl dd)

fun print_ndata(EMPTY_NDATA) = ()
  | print_ndata(NDATA(s)) = (print_string "NDATA ";                               
  			     print_string s)

(* entity declaration *)
fun print_entity_decl (GEDECL(s,ed)) = 
   (print_string ("<!ENTITY "^s^" "); 
    (case ed of
       GSTR_VAL(str) => print_quotes(str)
     | EXTERNAL_GENT(e_id, ndata) => print_external_id(e_id);
                                     print_string " ";      
                                     print_ndata(ndata)     
     );
     print_string " >\n")
  | print_entity_decl (PEDECL (s, pe)) = 
   (print_string ("<!ENTITY %"^s^" "); 
    (case pe of
       PSTR_VAL(str) => print_quotes(str)
     | EXTERNAL_PENT(e_id) => print_external_id(e_id)
     );
     print_string " >\n")

(* markups *)
fun print_markup  (ELEMENT_DECL (s, csp)) = 
	(print_string ("<!ELEMENT "^s^" ") ; print_content_spec csp ; print_string " >\n")
  | print_markup (ATTLIST_DECL (s1, alist)) =
        (print_string ("<!ATTLIST "^s1^"\n    "); 
         print_list print_att alist "\n    " ; 
         print_string " >\n")
  | print_markup (ENTITY_DECL e) = print_entity_decl e 
  | print_markup (NOTATION_DECL (s , pnd)) = 
	(print_string ("<!NOTATION "^s^" "); 
	 (case pnd of
		EXTERNAL_ID_NOTATION  exid => print_external_id exid
		| PUBID s  => (print_string "PUBLIC "; print_quotes s)
	 );
	print_string " >\n")
  | print_markup (PI_MARKUP t) = print_pi t
  | print_markup (COMMENT_MARKUP s ) = print_string (s^"\n")

(* internal type declaration *)
fun print_inbody (MARKUP m) = print_markup m

(* document type *)
fun print_doctype_miscs EMPTY_ODTDM = ()
  | print_doctype_miscs (ODTDM ((n,ex,inb), ms2)) = let 
      fun print_ext_header() = 
          print_string "<!-- this is included by ext dtd definition : start -->\n"
      fun print_ext_tailer() = 
          print_string "<!-- this is included by ext dtd definition : end -->\n"          
    in
      print_string ("<!DOCTYPE "^n^" ");
      (case (ex,inb) of 
         (EMPTY_EXTERNALID,EMPTY_INTERNAL_DECL) => ()
       | (EXTERNALID(eid,ext_subset),EMPTY_INTERNAL_DECL) =>
          print_external_id eid;
          print_string "\n[\n";
          print_ext_header();
          print_ast ext_subset;
          print_ext_tailer();
          print_string "\n]\n"
       | (EMPTY_EXTERNALID,INTERNAL_DECL_LIST inblist) =>
          print_string "\n[\n";
          print_list print_inbody inblist "\n";
          print_string "\n]\n"
       | (EXTERNALID(eid,ext_subset),INTERNAL_DECL_LIST inblist) =>
          print_external_id eid;
          print_string "\n[\n";
          print_list print_inbody inblist "\n";
          print_ext_header();
          print_ast ext_subset;
          print_ext_tailer();
          print_string "\n]\n"
        );
        print_list print_misc ms2 " ";
	print_string ">\n"
     end
	
(* prolog *)
and print_prolog (x, mlist, dm)  = 
	(print_xml_decl x; 
 	 print_list print_misc mlist " "; 
	 print_doctype_miscs dm)

(* element *)
and print_element (EMPTY_ELEMENT_TAG (s1, spairlist)) = 
	(print_string ("<"^ s1^" ");
	print_list (printpair print_string print_quotes "=") spairlist ", ";
	print_string "/>\n")
| print_element (FULL_ELEMENT (s2, spairlist, contentlist, s3)) = 
	(print_string ("<"^s2^" ");
	 print_list (printpair print_string print_quotes "=") spairlist "  ";
	 print_string ">\n ";
	 print_list print_content contentlist "\n";
	 print_string "\n </"; print_string s3; print_string ">\n")

(* content *)
and print_content (ELEMENT_CNT e ) =print_element e
  | print_content (CHARDATA_CNT s1) = print_string s1
  | print_content (CDATA_CNT s4) = print_string ("<![CDATA[ "^s4^"]]>")
  | print_content (PI_CNT t ) = print_pi t
  | print_content (COMMENT_CNT s7 ) = print_string (s7^"\n")

(* external_subset_declaration *)
and print_ex_subset_decl_elt (MARKUP_EXT m) = print_markup m
  | print_ex_subset_decl_elt (CONDITIONALSECT (s1, esd)) =
        (print_string "<![" ; 
	 (case s1 of 
		INCLUDE => print_string "INCLUDE" | IGNORE => print_string "IGNORE");
	 print_string "[";
	 print_ex_subset_decl esd;
	 print_string "]]>")
  | print_ex_subset_decl_elt (PEREF_EXT s2) = print_string ("%"^s2^";")
and print_ex_subset_decl eslist = print_list print_ex_subset_decl_elt  eslist "\n"

(* START POINT : root of ast *)
and print_ast (DOCUMENT (p, e, m)) = 
	(print_prolog p ; print_element e ; print_list print_misc m "")
  | print_ast (EXT_PARSEDENT (td, c)) =  (print_textdecl td; print_list print_content c "\n")
  | print_ast (EXT_PE (td, es)) =  (print_textdecl td; 
	print_ex_subset_decl es )

and print_root_elm_only(DOCUMENT (p, e, m)) = 
	(print_element e; print_list print_misc m "")
  | print_root_elm_only(EXT_PARSEDENT (td, c)) =  
        (print_string "this is external parsed entity \n")
  | print_root_elm_only(EXT_PE (td, es)) =  
	(print_string "this is external parameter entity \n")

end
