(*********************************************************************)
(*                erros.n                                            *)	
(*-------------------------------------------------------------------*)
(*  by Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(*  error processing                                                 *)
(*********************************************************************)

structure Nxml_errors = struct

open Global

type t_parser_error = 
  Not_allowed of (string * string * int )
| Illegal of  (string * string * int * string )
| Mismatch of (string * string * int)
| TagMismatch of (string * string * int)
| Bad_declaration of (string * string * int )
| Other_errors of (string * int)

exception XMLParserError of t_parser_error

fun  print_error_msg' (Illegal (s1,s2,ln, s4) ) = 
	(print_string "line "; print_int ln ; print_string (" : Illegal '"^s1^"' in "^s2^" : "^s4^"\n" )  )
   | print_error_msg' (Not_allowed (s1,s2,ln)) = 
	(print_string "line "; print_int ln ; print_string (" : '"^s1^"' is not allowed in "^s2^"\n")  )
   | print_error_msg' (Mismatch (s1,s2,ln) ) = 
	(print_string "line "; print_int ln ; print_string (" : '"^s1^"' is not found in "^s2^"\n") )
   | print_error_msg' (TagMismatch (s1,s2,ln) ) = 
	(print_string "line "; print_int ln ; print_string (" : Begining Tag '"^s1^"' is not match with ending Tag '"^s2^"'\n") )
   | print_error_msg' (Bad_declaration (s1, s2, ln)) = 
	(print_string "line "; print_int ln ; print_string (" : '"^s1^"' in "^s2^"\n") )
   | print_error_msg' (Other_errors (s,ln) ) = 
	(print_string "line "; print_int ln ; print_string (" : "^s^"\n")  )


fun print_error_msg(kind) = let
  fun print_current() = let 
    val uri = Global.current_URI()
  in
    print_string "Error in :"; 
    print_string uri; print_newline()
  end
in
  print_current();
  print_error_msg'(kind)
end

end
