(*********************************************************************)
(*                global.n                                           *)	
(*-------------------------------------------------------------------*)
(*  by Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(*  global variables and functions definition                        *)
(*   - debug tag                                                     *)
(*   - line handling                                                 *)
(*   - lexbuffer stack management                                    *)
(*   - uri stack management (for relative uri processing)            *)
(*********************************************************************)

structure Global = 
struct

val debug_tag = ref false

exception Exn_Global of string

(* line number *)
val line_no = ref 1
fun inc_line_no d = (line_no := (!line_no) + d)
fun count_line_no s =  
	let fun count_line_no_from s i =
		let val pos1 = (String.index_from s i '\n')
		in (count_line_no_from s (pos1+1) ) + 1 
		end handle Not_found => 0
	in count_line_no_from s 0 end

fun set_line_no s = inc_line_no (count_line_no s)

(**********************************************************)
(*  Absolute URI stack                                    *)
(**********************************************************)               
val uri_stack = ref ["file:."]
fun top_of_URI_stack() = List.hd(!uri_stack) 
                         handle _ => raise Exn_Global("empty uri stack")
// current_URI : used by System.string_from_system_literal
fun current_URI() = 
    if (!debug_tag) then 
      (print_string "\ncurrent_uri_stack :\n";  
      List.iter (fn x => print_string x; print_newline()) !uri_stack);
    top_of_URI_stack()
// pop_off : called by Lexer when it meet a EOF (and when it's external)
fun pop_off_a_URI() = 
   ((uri_stack := List.tl(!uri_stack))
    handle _ => raise Exn_Global("fail to pop off a uri stack : length="^(string_of_int (List.length !uri_stack))))
// push : by System.string_from_system_literal
fun push_a_URI(uri) = uri_stack := uri::(!uri_stack)


(**********************************************************)
(*   Global Lexbuff (set dynamically by lexer and parser  *)
(**********************************************************)
type lexbuf_stack = (Lexing.lexbuf * bool(* is_external *) * int (*line no of past buff*)) list ref
val lexbuf_stack : lexbuf_stack  = ref []

fun top_of_lexbuf_stack() = List.hd(!lexbuf_stack) 

// used by lexer every time it called
fun current_lexbuf() = let 
  val (buff,_,_) = top_of_lexbuf_stack()
in buff end        

// popped by lexer every time it meets Eof exception when it's a last lexbuf
fun pop_off_a_lexbuf() = let 
  val (_,_,lineno) = top_of_lexbuf_stack() 
in
  lexbuf_stack := List.tl(!lexbuf_stack);
  lineno
end

// pushed by parser every time it meets a well-formed/valid entity Reference
fun push_a_lexbuf(lexbuf,b,l_no) = 
    (line_no := 1;
     lexbuf_stack := (lexbuf,b,l_no)::(!lexbuf_stack))
fun is_current_lexbuf_external() = let
  val (_,b,_) = top_of_lexbuf_stack()
in b end

end