(*********************************************************************)
(*                notation.n                                         *)	
(*-------------------------------------------------------------------*)
(*  by Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(*  Implement notation related processing                            *)
(*********************************************************************)


signature NOTATION_PROCESSING =
sig  
  type notation = N_PUB of string 
          | N_PUBLIC of string * string 
          | N_SYSTEM of string 
      
  val notation_env_insert : string  // notation name
                         -> notation // system literal 
                         -> unit 
  val notation_env_lookup : string  // notation name
                         -> notation  // system literal
  val notation_notity : string -> unit
  val notation_exists : string -> bool
                           
end  

structure Notation_processing : NOTATION_PROCESSING  =
struct

open Global
open Nxml_errors

type notation = N_PUB of string 
          | N_PUBLIC of string * string 
          | N_SYSTEM of string 

// notation environment 
type id = string
type system_literal = string
type n_env = (id * notation) list

val emptyNE : n_env = [] 

val n_env = ref emptyNE
fun notation_env_insert id notation = n_env := (id,notation)::(!n_env)

fun notation_env_lookup id = 
let 
  val (_,notation) = List.find (fn (x,_) => x=id) !n_env 
    		        handle Not_found => 
			  raise XMLParserError(
				   Other_errors ("undefiend notation name reference:"^id,!line_no))
    		      | x => raise x	
in notation end

fun notation_notity id = ()
fun notation_exists id = List.exists (fn (x,_) => x=id) !n_env

end  