%{
(*********************************************************************)
(*                Parser.ny                                          *)	
(*-------------------------------------------------------------------*)
(*  Initiatial development by Eun-sun Cho                            *)
(*  Finalized by Hyun-Goo Kang, ROPAS                                *)
(*-------------------------------------------------------------------*)
(*  http://ropas.kaist.ac.kr/~hgkang                                 *)
(*********************************************************************)

(*****************************************)
(*         To do List : 7/31             *)
(*****************************************)
(*
  - validation  
  - external entity's textdecl
  - more kind of uri, more relative uri
  - Conditional Section : it should be possible in external parameter entity only
  - public literal support 
  - encoding : assumed as ISO-8859/1 (2 byte encoding.. : real challenge)
  - language identifying
*)

open Nxml_ast
open Nxml_errors
open Entity_proccesing
open Notation_processing
open Global

type doc_kind = DT_DOCUMENT 
              | DT_EXT_PARSED_ENT 
              | DT_EXT_PE 

%}  

/*********************************************************/
/*                 Declaration Section                   */
/*********************************************************/

/*---------------------------------------------------------*/
/*            Tokens : Terminal_Symbols                    */
/*---------------------------------------------------------*/
%token T_Public T_System
%token Bar Comma Conditional_begin Conditional_end Conditional_body Dtd_end LBRACK
%token Decl_attlist Decl_element Decl_entity Decl_notation Decl_rangle 
%token Doctype Doctype_rangle Dtd_begin Eof Eq 
%token Lparen Ignore Implied Percent Pcdata Plus Qmark Rangle Rangle_empty 
%token Required Rparen RparenPlus RparenQmark RparenStar Star
%token XMLDecl_begin XMLDecl_end PI_NULL_begin Pcdata Implied Required Fixed 
%token <string> Attval Comment Cdata 
%token <string> CharData ERef Name Nametoken PERef 
%token <string> Tag_beg Tag_end Unparsed_string Version 
%token <string> Encoding Standalone PI_end PI_begin
%token <int> CRef

/*---------------------------------------------------------*/
/*     %type <produced_absyn_type> Non_Terminal_Symbol     */
/*---------------------------------------------------------*/
%type <Nxml_ast.t_entry_point> ext_parsed_ent ext_pe ext_pe_
%type <Nxml_ast.t_entry_point> ext_document 
%type <t_prolog> prolog
%type <t_opt_doctype_decl_miscs> opt_doctype_decl_miscs
%type <t_opt_xml_decl> opt_xml_decl xml_decl
%type <t_opt_encoding> opt_encoding_decl
%type <t_opt_standalone> opt_SD_decl
%type <t_doctype_decl> doctype_decl
%type <t_opt_external_id * t_opt_internal_decl_list> doctype_body
%type <t_internal_decl_body list> internal_decl internal_decl_body_list
%type <unit> declsep declsep2
%type <t_markup_decl>  markup_decl element_decl attlist_decl
%type <t_content_spec> content_spec
%type <t_pre_children * t_marks> children
%type <t_marks> star_plus_qmarks
%type <t_cp> cp
%type <t_cp list> choice choice_body seq seq_body
%type <t_mixed> mixed mixed_end
%type <string list> mixed_name_list 
%type <string list> opt_notation notation_type_body enumeration_body
%type <(string * t_att_type * t_default_decl) list> att_def_list
%type <string * t_att_type * t_default_decl> att_def
%type <t_att_type> att_type
%type <t_default_decl> default_decl
%type <string> att_value
%type <string> tag_end name_token
%type <t_entity_decl>entity_decl ge_decl pe_decl 
%type <t_entity_def> entity_def 
%type <t_pe_def> pe_def 
%type <string>entity_value
%type <t_opt_ndata> opt_ndata
%type <t_external_id> external_id system_external_id public_external_id external_id_sep
%type <string * t_post_notation_decl> notation_decl
%type <t_post_notation_decl> notation_body
%type <t_pi> pi
%type <t_misc list> miscs
%type <t_element> element
%type <string * ((string *string) list)> empty_element_tag
%type <(string *string) list> attribute_list
%type < string * ((string * string) list) * t_content list * string> full_element
%type <string * ((string * string) list )> tag_start
%type <t_content list > content
%type <t_opt_text_decl > opt_text_decl text_decl
%type <t_opt_version > opt_version
%type <t_ext_subset_decl> ext_subset_decl
%type <t_include * t_ext_subset_decl> conditional_sect

%start ext_document
%start ext_parsed_ent
%start ext_pe

%%

/*********************************************************/
/*                    Rule Section                       */
/*********************************************************/


/********************************************************************/
/*   see the usage of declsep, declsep2, external_id_sep rule       */
/*------------------------------------------------------------------*/
/*    - it's for inserting entity string into lexbuf                */
/*    - stangely it works :-)                                       */
/*------------------------------------------------------------------*/
/*  - external_id_sep, declsep, declsep2                            */
/*  - push the string into the top of lexbuf stack                  */
/*  - top of lexbuf stack is the string to parse next               */
/*  - so, pushed string will be parsed next                         */    
/*  - this production is dealed as if it is a hole in the grammar   */
/*  - so, after this production, inserted string should be matched  */
/*    in a production in original production in XML spec            */
/*  - so, we should insert productions for the hole                 */
/********************************************************************/

/* this rule help an external document type definition to be joined in parsing */
/*    using lexbuf handling (an ad hoc approach, but it works well             */
external_id_sep : /* external_id */
       T_System Unparsed_string 
       { 
        let 
          val s = $2
          val external_id = SYSTEM(s)
          val absolute_uri = System_Processing.absolute_path (current_URI()) s 
          val _ = push_a_URI(absolute_uri)
          val str = System_Processing.string_from_system_literal(absolute_uri)
          val new_lexbuf = Lexing.from_string str                  
        in                 
          push_a_lexbuf(new_lexbuf, true, !line_no);
          external_id
        end
       } 
       | T_Public Unparsed_string Unparsed_string
       { 
        let 
          val s = $3
          val external_id = PUBLIC($2,s)
          val absolute_uri = System_Processing.absolute_path (current_URI()) s 
          val _ = push_a_URI(absolute_uri)
          val str = System_Processing.string_from_system_literal(absolute_uri)
          val new_lexbuf = Lexing.from_string str                  
        in 
          push_a_lexbuf(new_lexbuf, true, !line_no);
          external_id
        end
       }        
          
declsep : PERef 
	  {
	    // external parsed pe + internal parsed pe
            // (this production shouldn't appear in inner part of markupdecl)
	  let  	    
	    val p_entity = pe_entity_env_lookup $1	    
	    val is_external = pe_is_external p_entity
	    val _ = if (is_external) then 
	            let val uri = pe_get_uri(p_entity)
	            in  if (!Global.debug_tag) then 
	                   print_string ("pushing a uri "^uri);
	                push_a_URI(uri) 
	            end	        
	    val the_string = pe_extract_string p_entity	    
	    val _ = Global.current_URI()
	    val new_lexbuf = Lexing.from_string the_string
	  in push_a_lexbuf(new_lexbuf, (pe_is_external(p_entity)), !line_no) end
          }  /*{28a}*/                

declsep2 : PERef 
          {
            // only for internal parsed pe
            // (it can appear in inner part of markupdecl production as long as it can be parsed properly)
	  let  	    
	    val p_entity = pe_entity_env_lookup $1
	    val _ = if (pe_is_external p_entity) then 
	            raise XMLParserError(
	                    Other_errors("an external pe ref can't appear in inner part of"
	                             ^ " markupdecl production",!line_no))	            
	    val the_string = pe_extract_string p_entity	    
	    val new_lexbuf = Lexing.from_string the_string
	  in 
	     if (!Global.debug_tag) then 
	        print_string "pushing a internal pe";
	     push_a_lexbuf(new_lexbuf,false, !line_no) 
	  end            
	  }




/* START SYMBOL
   {01} document: parses a complete XML document 
                  (i.e. containing a <!DOCTYPE..> and an element)
*/
ext_document: prolog  element  miscs Eof  { DOCUMENT($1, $2,List.rev($3)) } 

/* {22} prolog */
prolog : opt_xml_decl miscs opt_doctype_decl_miscs 
	{ ($1, List.rev($2), $3) }

opt_doctype_decl_miscs : 
        { EMPTY_ODTDM }
	| doctype_decl miscs 
	{ ODTDM($1, List.rev($2))}

opt_xml_decl : 
        {EMPTY_XML_DECL}
	| xml_decl {$1}

/* {23} XMLDecl */
xml_decl : XMLDecl_begin Version opt_encoding_decl opt_SD_decl XMLDecl_end
	{XML_DECL (VERSION($2), $3, $4)}

/* {80} EncodingDecl */
opt_encoding_decl : {EMPTY_ENCODING}
	| Encoding {ENCODING($1)} /* warning in validation */

/* {32} SDDecl */
opt_SD_decl : {EMPTY_STANDALONE }
	| Standalone 
	{ 	case $1 of 
			"yes" => YES_STANDALONE
                      | "no" => NO_STANDALONE 
		      | _ => raise XMLParserError(Illegal($1, 
					"stand alone parameter",!line_no,"")) 
	}

/* {28} doctypedecl */
doctype_decl : Doctype Name doctype_body Doctype_rangle 
	{ ($2, fst $3, snd $3) }

doctype_body : 
        external_id_sep ext_pe_ opt_internal_decl  {(EXTERNALID($1,$2), $3)}      
      | opt_internal_decl {(EMPTY_EXTERNALID,$1)}
   
opt_internal_decl : {EMPTY_INTERNAL_DECL}
	| internal_decl {INTERNAL_DECL_LIST($1)}

internal_decl : LBRACK internal_decl_body_list Dtd_end 
	{
	List.rev($2)
	}

internal_decl_body_list : {nil}
	| internal_decl_body_list markup_decl {MARKUP($2)::$1}
	| internal_decl_body_list declsep {$1}

/* {29} markupdecl */
markup_decl :
	  element_decl 	{$1}
	| attlist_decl 	{$1}
	| entity_decl 	{ENTITY_DECL($1)}
	| notation_decl { NOTATION_DECL($1) }
	| pi 			{ PI_MARKUP($1) }
	| Comment 		{COMMENT_MARKUP($1)}	

/* {45} elementdecl */
element_decl : Decl_element Name content_spec Decl_rangle 
	{ELEMENT_DECL ($2, $3)}
	| Decl_element declsep2 Name content_spec Decl_rangle
	{ELEMENT_DECL ($3, $4)}	
  
/* {46} contentspec */
content_spec : Name 
	{ case $1 of 
		"EMPTY" => EMPTY_CS
		| "ANY" => ANY_CS 
		| _ => raise XMLParserError(Illegal($1, "content spec",!line_no,"")) }
	| mixed 	        {MIXED($1)}		
	| children              {CHILDREN($1)}
	| declsep2 Name 
	{ case $2 of 
		"EMPTY" => EMPTY_CS
		| "ANY" => ANY_CS 
		| _ => raise XMLParserError(Illegal($2, "content spec",!line_no,"")) }	
	| declsep2 mixed         {MIXED($2)}
	| declsep2 children 	{CHILDREN($2)}

/* {47} children */
children : choice star_plus_qmarks  {(CHOICE($1),$2)}
	| seq star_plus_qmarks  		{(SEQ($1),$2)}

star_plus_qmarks : {NO_MARK}
	| Star  {STAR}
	| Plus  {PLUS}
	| Qmark {QMARK}

/* {49} choice */
choice : Lparen choice_body Rparen 
	{List.rev($2)}

choice_body : cp  {$1::nil}
	| choice_body Bar cp { $3::$1}

/* {48} cp */
cp : 	Name star_plus_qmarks 	{(NAME_CP($1),$2)}
	| choice star_plus_qmarks 	{(CHOICE_CP($1),$2) }
	| seq star_plus_qmarks 		{(SEQ_CP($1), $2)}
	/* following 3 are for peref expansion */
	| declsep2 Name	 star_plus_qmarks {(NAME_CP($2),$3)} 
	| declsep2 choice star_plus_qmarks  {(CHOICE_CP($2),$3) }
	| declsep2 seq star_plus_qmarks     {(SEQ_CP($2),$3) }


/* {50} seq */
seq : Lparen seq_body Rparen {List.rev($2)}

seq_body : cp Comma cp {
		// According to the original XML definition, 
		// this rule must be consist of only 'cp', instead of 'cp Comma cp'.
		// However it raises a reduce-reduce conflict with 'choice_body' rule.
		// In order to avoid this problem 
		// I changed 'seq_body' to have at least two cp's 
		$3::$1::nil
	}  
	| seq_body Comma cp  {$3::$1}

/* {51} Mixed */
mixed : Lparen Pcdata mixed_end 
        { 
          let 
            fun duplicate_name_check([]) = false
	      | duplicate_name_check(id::tl) = 
	         (List.exists (fn(id')=>(id'=id)) tl) 
	         || duplicate_name_check(tl)
	  in
	    (case $3 of 
	        MIXED_ONLY => $3
	      | MIXED_STAR(l) => 
	        if (duplicate_name_check l) then 
	          raise XMLParserError(Other_errors("No Duplicate Types: "
	          	        ^ " The same name must not appear more than once "
	        	        ^ "in a single mixed context decl",!line_no))
	        else
	          $3)
	  end        	        
        }

mixed_end : Rparen  { MIXED_ONLY }
	| Rparen  Star 	{ MIXED_STAR (nil) }
	| mixed_name_list Rparen Star { MIXED_STAR(List.rev($1))}

mixed_name_list : Bar Name 		{$2::nil} 
        | Bar declsep2 Name 		{$3::nil} 
        | mixed_name_list Bar Name 	{$3::$1}
	| mixed_name_list Bar declsep2 Name 	{$4::$1}

/* {52} AttlistDecl */
attlist_decl : Decl_attlist Name att_def_list Decl_rangle 
	       {ATTLIST_DECL($2,List.rev($3))}
	     | Decl_attlist declsep2 Name att_def_list Decl_rangle 
	       {ATTLIST_DECL(($3),List.rev($4))}  

att_def_list :  {nil} 
              | att_def_list att_def {$2::$1}

/* {53} AttDef */
att_def : Name att_type default_decl {($1,$2,$3)}
        | declsep2 Name att_type default_decl {($2,$3,$4)}
            
/* {54} AttType  = 
   {55} string_type : CDATA
   {56} tokenized_type :{ID | IDREF | IDREFS | 
			ENTITY | ENTITIES | NMTOKEN | NMTOKENS}
   {57} enumerated_type: 
   {58} natation_type : NOTATION (Name | Name | Name )
   {59} enumeration : (Nametoken | Nametoken | Nametoken )
*/
att_type : 
	  Name opt_notation 
	{
	  case $2 of 
	    nil => 
	      (case $1 of 
		 "CDATA" => CDATA
	       | "ID" => ID // this is value uniquely idenify the elements which bear them
	       | "IDREF" => IDREF
	       | "IDREFS" => IDREFS
	       | "ENTITY" => ENTITY
	       | "ENTITIES" => ENTITIES
	       | "NMTOKEN" => NMTOKEN
	       | "NMTOKENS" => NMTOKENS
	       | _ => raise XMLParserError(Illegal($1, "type name",!line_no,
		"The name should be one of the CDATA/ID/IDREF(S)/ENTITY(S)/NMTOKEN(S)")) 
	       )
	   | a::alist => 
	       (case $1 of 
	          "NOTATION" => NOTATION(a::alist)
	       | _  => raise XMLParserError(Illegal($1, "type name",!line_no,"illgal notation type"))
	       )  
	}
        | declsep2 Name opt_notation 
	{
	  case $3 of 
	    nil => 
	      (case $2 of 
		 "CDATA" => CDATA
	       | "ID" => ID // this is value uniquely idenify the elements which bear them
	       | "IDREF" => IDREF
	       | "IDREFS" => IDREFS
	       | "ENTITY" => ENTITY
	       | "ENTITIES" => ENTITIES
	       | "NMTOKEN" => NMTOKEN
	       | "NMTOKENS" => NMTOKENS
	       | _ => raise XMLParserError(Illegal($2, "type name",!line_no,
		"The name should be one of the CDATA/ID/IDREF(S)/ENTITY(S)/NMTOKEN(S)")) 
	       )
	   | a::alist => 
	       (case $2 of 
	          "NOTATION" => NOTATION(a::alist)
	       | _  => raise XMLParserError(Illegal($2, "type name",!line_no,"illgal notation type"))
	       )  	   
	}
	| declsep2 Lparen enumeration_body Rparen 
	{ ENUMERATION(List.rev($3)) }	
	| Lparen enumeration_body Rparen 
	{ ENUMERATION(List.rev($2)) }

opt_notation :  				{nil}   /* string_type, tokenized_type */
	| Lparen notation_type_body Rparen 	{List.rev($2)}

/* {58} NotationType */
notation_type_body : 
          Name 	
        { if (notation_exists $1) then
             $1::nil
          else   
             raise XMLParserError(
	           Other_errors("undefined notation reference in Notation Attributes: "
	                         ^ $1,!line_no))
        }	                         
	| notation_type_body Bar Name 	
	{ if (notation_exists $3) then
             $3::$1
          else   
             raise XMLParserError(
	           Other_errors("undefined notation reference in Notation Attributes: "
	                         ^ $3,!line_no))
        }	                      	

/* {08} Nmtokens */
name_token : Nametoken {$1}
	| Name {$1 
		(* Name is introduced for Nametoken because 'Name' is subset  of 'Nametoken'
		and the lexer returns 'Name' instead of 'Nametoken' when it matches 'Name' *) }

/* {59} Enumeration */
enumeration_body : name_token  {$1::nil}
	| declsep2 name_token  {$2::nil}
	| enumeration_body Bar name_token {$3::$1}
	| enumeration_body Bar declsep2 name_token {$4::$1}

/* {60} DefaultDecl */
default_decl : Required  	{REQUIRED}
	| Implied  			{IMPLIED}
	| Fixed att_value 		{FIXED($2)}
	| att_value 			{ATTVALUE($1)}

/* {10} AttValue */
att_value : Unparsed_string { normalize_attr_value($1) }

/* {70} EntityDecl */
entity_decl :  Decl_entity ge_decl Decl_rangle 		{$2}  /* general entity */
             | Decl_entity pe_decl Decl_rangle 	 	{$2}  /* parameter entity */

/* {71} GEDecl */
ge_decl : Name entity_def 		
        {
         let 
           val ge_def = $2         
         in 
           (case ge_def of 
             GSTR_VAL(val_str) => ge_entity_env_insert $1 (G_STR(val_str))
           | EXTERNAL_GENT(SYSTEM(s),EMPTY_NDATA) =>  
               ge_entity_env_insert $1 (G_REF(SYSTEM_REF(s),EMPTY_NOTATION))
           | EXTERNAL_GENT(SYSTEM(s),NDATA(ndata)) =>  
               if (notation_exists ndata) then
                 ge_entity_env_insert $1 (G_REF(SYSTEM_REF(s),N(ndata)))
               else
                 raise XMLParserError(
	                 Other_errors("undefined notation reference in ge definition: "
	                             ^ ndata,!line_no))
           | EXTERNAL_GENT(PUBLIC(s1,s2),EMPTY_NDATA) => 
               ge_entity_env_insert $1 (G_REF(PUBLIC_REF(s1,s2),EMPTY_NOTATION))
           | EXTERNAL_GENT(PUBLIC(s1,s2),NDATA(ndata)) =>
               if (notation_exists ndata) then
                 ge_entity_env_insert $1 (G_REF(PUBLIC_REF(s1,s2),N(ndata)))
               else
                 raise XMLParserError(
	                 Other_errors("undefined notation reference in ge definition: "
	                             ^ ndata,!line_no))
           );
            GEDECL($1,$2)
          end                  
        }

/* {72} PEDecl */
pe_decl : Percent Name pe_def 	
        {
         let val pe_def = $3
         in 
           (case pe_def of 
             PSTR_VAL(val_str) => 
               pe_entity_env_insert $2 (P_STR(val_str))
           | EXTERNAL_PENT(SYSTEM(s)) => 
               pe_entity_env_insert $2 (P_REF(SYSTEM_REF(s)))
           | EXTERNAL_PENT(PUBLIC(s1,s2)) => 
               pe_entity_env_insert $2  (P_REF(PUBLIC_REF(s1,s2)))
            );
            PEDECL($2,$3)
          end  
         }

/* {73} EntityDef */
entity_def : 
          entity_value 		{GSTR_VAL($1)}
        | declsep2 entity_value {GSTR_VAL($2)}  
	| external_id opt_ndata 	{EXTERNAL_GENT($1,$2)}

/* {74} PEDef */
pe_def: entity_value 			{PSTR_VAL($1)}
      | external_id			{EXTERNAL_PENT($1)}
      | declsep2 entity_value 		{PSTR_VAL($2)}
      | declsep2 external_id		{EXTERNAL_PENT($2)}

/* {74} NDataDecl */
opt_ndata : {EMPTY_NDATA}
        | Name Name 
	{ if ($1 = "NDATA") then NDATA($2)
		else  raise XMLParserError(Illegal($1, "NDataDecl", 
			!line_no, "keyword 'NDATA' is expected" )) }

/* {09} EntityValue */
entity_value : Unparsed_string {normalize_entity_value($1)}

/* {75} ExternalId  =
 "SYSTEM" SystemLiteral{11}
 | "PUBLIC" PubidLiteral{12} SystemLiteral
*/
external_id : 
	system_external_id {$1} | public_external_id {$1}

system_external_id :  
	T_System Unparsed_string  
	{
	  SYSTEM($2)
	}

public_external_id :
	T_Public Unparsed_string Unparsed_string  { // first unparsed should be a public char seq
		 PUBLIC($2,$3)
		}

/* {82} NotationDecl  
  = '<!NOTATION' S Name S (ExternalID | PublicID ) S? '>' */
notation_decl : Decl_notation  Name notation_body Decl_rangle 
        {
          let 
            val n_body = $3
          in
            (case n_body of 
                EXTERNAL_ID_NOTATION(SYSTEM(s)) => 
                   notation_env_insert $2 (N_SYSTEM s) 
              | EXTERNAL_ID_NOTATION(PUBLIC(s1,s2)) =>
                   notation_env_insert $2 (N_PUBLIC(s1,s2))
              | PUBID(s) =>
                   notation_env_insert $2 (N_PUB s));
            ($2, $3)
          end  
        }

notation_body : 
	T_System Unparsed_string {EXTERNAL_ID_NOTATION(SYSTEM($2))}
      | T_Public Unparsed_string {PUBID($2)}
      | T_Public Unparsed_string Unparsed_string { EXTERNAL_ID_NOTATION(PUBLIC($2,$3))}   

/* {16} PI  */ 
pi : PI_begin PI_end 	{ PI_WITH_TARGET($1,$2)}
| PI_NULL_begin PI_end 	{ PI_WO_TARGET($2)}

/* {27} Misc */
miscs : 			{nil}
	| miscs Comment {COMMENT($2)::$1}
	| miscs pi 		{PI($2)::$1}

/* {39} element */
element : empty_element_tag  	{EMPTY_ELEMENT_TAG($1)}
	| full_element 		{FULL_ELEMENT($1)}


/* {44} EmptyElemTag */
empty_element_tag : Tag_beg  attribute_list Rangle_empty {($1,List.rev($2))}

attribute_list  : 	  			{nil} 
	| attribute_list Name Eq att_value {($2, $4)::$1}

/* if att_value has type entity then "entity_name" should be notified to application 
   directed by proper ndata on that entity 
   => this should be implemented in validation phase
*/

full_element : tag_start content tag_end 
	{ if ((fst $1) = $3) then (fst $1, snd $1,List.rev($2), $3)
		else raise XMLParserError(TagMismatch(fst $1, $3, !line_no)  ) }

/* {40} STag */
tag_start : Tag_beg attribute_list Rangle 
	    {
	      let 
	        val l = List.rev($2)
	        fun duplicate_att_name_check([]) = false
	          | duplicate_att_name_check((id,att_val)::tl) = 
	            (List.exists (fn(id',_)=>(id'=id)) tl) 
	            || duplicate_att_name_check(tl)
	               
	        val _ = if (duplicate_att_name_check l) then 
	        	raise XMLParserError(Other_errors("Unique Att spec:"
	        	      ^ " no attibutename may appear more than once"
	        	      ^ " in the same start-tag or empty_element tag",!line_no))
	      in 
	      	($1,l)
	      end
	    }

/* {43} content */
content :  {nil}
	| content element 	{(ELEMENT_CNT $2)::$1}
	| content CharData 	{(CHARDATA_CNT $2)::$1}
	| content CRef 		
	{
	  let 
	    val num = $2
	  in
            if (num > 255) then // fix
              (print_string ("Warning: this processor don't support 2 byte character(num_str="^(string_of_int num)^")\n");
               print_string ("         substituting it to ?\n");
               CHARDATA_CNT("?")::$1)               
            else  
              CHARDATA_CNT(String.make 1 (Char.chr(num)))::$1
	  end
	}
	| content ERef 		
	{ let
	    val g_entity = ge_entity_env_lookup $2
	  in
	    if (ge_is_unparsed g_entity) then
	      raise XMLParserError(
	             Other_errors("unparsed entity ref is forbidden in contents",!line_no))
	    else
	    let
              val is_external = ge_is_external g_entity
	      val _ = if (is_external) then 
	              let val uri = ge_get_uri(g_entity)
	              in push_a_URI(uri) end	         
	      val the_string =  ge_extract_string(g_entity)
	      // ⼭ external̸ textdecl 
	      //  Ʈ Եȴ.
	      val new_lexbuf = Lexing.from_string (the_string)
	    in
	      push_a_lexbuf(new_lexbuf,is_external,!line_no);
	      $1  // so EREF_CNT can't exist in absyn
	    end
	    // for external parsed entity : the ruls is "included if validating"
	    // for i didn't implemete validation yet, i always included it. 
	    // if there is any error you meet just a exception
	  end  
	 }
	| content Cdata  	{(CDATA_CNT $2)::$1}
	| content pi 		{(PI_CNT $2)::$1 }
	| content Comment 	{(COMMENT_CNT $2)::$1}

/* {42} ETag */
tag_end : Tag_end Rangle {$1}

opt_text_decl : {EMPTY_TEXTDECL}
	| text_decl{$1}

/* {77} TextDecl */
text_decl : XMLDecl_begin  opt_version Encoding XMLDecl_end 
	{ case $2 of
		EMPTY_VERSION => TEXTDECL($3) 
		| VERSION t => TEXTDECL_WITHVERSION(t,$3) 
	}

opt_version : {EMPTY_VERSION}| Version{VERSION($1)}

/* {31} extSubsetDecl */
ext_subset_decl : {nil}
	| ext_subset_decl markup_decl /* scan_declaration */
	{ MARKUP_EXT($2)::$1}
	| ext_subset_decl conditional_sect /*scan declaration  */
	{ CONDITIONALSECT($2)::$1}
	| ext_subset_decl declsep  {$1}

/* {61} conditionalSect */
conditional_sect : Conditional_begin Name LBRACK ext_subset_decl Conditional_end
        { case  $2 of
	  "INCLUDE" => (INCLUDE, $4)
	| "IGNORE" => (IGNORE, $4)
	| _ => raise XMLParserError(Illegal($2, "conditional section declaration",
	      !line_no,  ": any keyword except INCLUDE or IGNORE is disallowed" )) }

/* START SYMBOL
 *
 * {78} extParsedEnt:     parses a external XML entities to be parsed 
 *			 but not containing a data type declation
 */
ext_parsed_ent : opt_text_decl content Eof 
	{ EXT_PARSEDENT($1,List.rev($2)) }

/* START SYMBOL
 * 
 * {89} extPE:       parses a external declaration of data type declaration
 */
ext_pe : opt_text_decl ext_subset_decl Eof 
	{EXT_PE($1,List.rev($2))}
ext_pe_ : opt_text_decl ext_subset_decl 
	{EXT_PE($1,List.rev($2))}	

%%

(*********************************************************)
(*                 Trailer Section                       *)
(*********************************************************)

(*---------------------------------------------------------*)
(*          Wrappers for each start_symbol(parser)         *)
(*---------------------------------------------------------*)
(* entry points of parser 				 					*)
(* wrapped with handling the parser_error with line number 	*)
(* {78} extParsedEnt *)
fun extParsedEnt f b = (ext_parsed_ent f b ) 
	handle Parse_error =>  raise XMLParserError(Other_errors("Syntax error in parsed entity ", !line_no )) 
	| x  => raise x
(* {79} extPE *)
fun extPE  f b = (ext_pe f b )
	handle Parse_error =>  raise XMLParserError(Other_errors("Syntax error in external parameter entities", !line_no )) 
	| x  => raise x
(* {01} document *)
fun document  f b = (ext_document  f b) 
	handle Parse_error =>  raise XMLParserError(Other_errors("Syntax error in external document", !line_no )) 
	| x  => raise x

(*
// main parser function 
fun parse lexfun uri option = let  
  val absolute= System_Processing.absolute_path (Global.current_URI()) uri
  val _ = Global.push_a_URI absolute 
  val ducument_string = System_Processing.string_from_system_literal absolute // it's normalized
  val lexbuf = Lexing.from_string ducument_string
in 
  case option of 
    DT_DOCUMENT => document lexfun lexbuf
  | DT_EXT_PARSED_ENT => extParsedEnt lexfun lexbuf
  | DT_EXT_PE => extPE lexfun lexbuf
end
*)

(*---------------------------------------------------------*)
(*          Pretty Printer for Debugging                   *)
(*---------------------------------------------------------*)
(* print each token for debugging *)
fun print_tok tok = 
(	case tok of
    Bar => print_string "(|)"
  | Comma => print_string "(,)"
  | T_System => print_string "SYSTEM"  
  | T_Public => print_string "PUBLIC"
  | Conditional_begin => print_string "(Conditional_begin)"
  | Conditional_end => print_string "(Conditional_end)"
  | Conditional_body => print_string "(Conditional_body)"
  | LBRACK => print_string "([)"
  | Dtd_end => print_string "(Dtd_end)"
  | Decl_attlist => print_string "(Decl_attlist)"
  | Decl_element => print_string "(Decl_element)"
  | Decl_entity => print_string "(Decl_entity)"
  | Decl_notation => print_string "(Decl_notation)"
  | Decl_rangle => print_string "(Decl_rangle)"
  | Doctype => print_string "(Doctype)"
  | Doctype_rangle => print_string "(Doctype_rangle)"
  | Dtd_begin => print_string "(Dtd_begin)"
  | Eof => print_string "(Eof)"
  | Eq => print_string "(Eq)"
  | Lparen => print_string "(Lparen)"
  | Ignore => print_string "(Ignore)"
  | Implied => print_string "(Implied)"
  | Percent => print_string "(Percent)"
  | Pcdata => print_string "(Pcdata)"
  | Plus => print_string "(Plus)"
  | Qmark => print_string "(Qmark)"
  | Rangle => print_string "(Rangle)"
  | Rangle_empty => print_string "(Rangle_empty)"
  | Required => print_string "(Required)"
  | Rparen => print_string "(Rparen)"
  | RparenPlus => print_string "(RparenPlus)"
  | RparenQmark => print_string "(RparenQmark)"
  | RparenStar => print_string "(RparenStar)"
  | Star => print_string "(Star)"
  | XMLDecl_begin => print_string "(XMLDecl_begin)"
  | XMLDecl_end => print_string "(XMLDecl_end)"
  | Fixed => print_string "(Fixed)"
  | Attval t => print_string ("(Attval:"^t^")")
  | Comment t =>  print_string ("(Comment:"^t^")")
  | Cdata  t => print_string ("(Cdata:"^t^")")
  | CharData  t => print_string ("(CharData:"^t^")")
  | ERef t =>print_string ("(ERef:"^t^")")
  | Name t => print_string ("(Name:"^t^")")
  | Nametoken t => print_string ("(Nametoken:"^t^")")
  | PERef t => print_string ("(PERef:"^t^")")
  | Tag_beg t => print_string ("(Tag_beg:"^t^")")
  | Tag_end t => print_string ("(Tag_end:"^t^")")
  | Unparsed_string t => print_string ("(Unparsed_string:"^t^")")
  | Version t => print_string ("(Version:"^t^")")
  | Encoding t => print_string ("(Encoding:"^t^")")
  | Standalone t => print_string ("(Standalone:"^t^")")
  | PI_end t => print_string ("(PI_end:"^t^")")
  | PI_begin t => print_string ("(PI_begin:"^t^")")
  | PI_NULL_begin => print_string ("(PI_NULL_begin)")
  | CRef t => (print_string "(CRef:"; print_int t; print_string ")")
)
