(*********************************************************************)
(*                test.n                                             *)	
(*-------------------------------------------------------------------*)
(*  by Hyun-Goo Kang                                                 *)
(*-------------------------------------------------------------------*)
(*  example use of the XML processor                                 *)
(*   - parsing then printing                                         *)
(*********************************************************************)

open NXML

val doct = ref DT_DOCUMENT
val validation_tag = ref false
val print_root_only = ref false

fun test uri = let  
  val absyn = parse uri !doct
in
  if (!validation_tag) then 
    validate(absyn);
  if (!print_root_only) then 
    Nxml_ast.print_root_elm_only(absyn)
  else
    Nxml_ast.print_ast absyn
  ; 
  print_newline()
end 

val usage = "\nUsage: go [-dnpgev] <uri or filename>\n\n  e.g) go 1.xml\n       go http://www.w3.org\n"
val optionlist = 
	[("-", Arg.Unit (fn () => doct:= DT_DOCUMENT),
	"  : <uri or filename> is a normal XML document to start"),
	("-n", Arg.Unit (fn () => doct:= DT_DOCUMENT),
	" : <uri or filename> is a normal XML document to start"),
	("-p", Arg.Unit (fn () => doct:= DT_EXT_PE),
	" : <uri or filename> is an external subset of data type definition"),
	("-g", Arg.Unit (fn () => doct:= DT_EXT_PARSED_ENT),
	" : <uri or filename> is an external parsed entity (with content)"),
	("-e", Arg.Unit (fn () => print_root_only := true),
	" : for printing root element only"),
	("-d", Arg.Unit (fn () => (set_debug_mode())),
	" : for debugging (for hgkang)"),
	("-v", Arg.Unit (fn () => validation_tag := true),
	" : for validation (not completed)")
	 ]

fun main() = 
  Format.set_formatter_out_channel stderr;
  Arg.parse optionlist test usage; 
  exit 0 

val _ = main()

